/*********************************************************************
 *
 * Copyright:
 *	1999-2000 MOTOROLA, INC. All Rights Reserved.  
 *  You are hereby granted a copyright license to use, modify, and
 *  distribute the SOFTWARE so long as this entire notice is
 *  retained without alteration in any modified and/or redistributed
 *  versions, and that such modified versions are clearly identified
 *  as such. No licenses are granted by implication, estoppel or
 *  otherwise under any patents or trademarks of Motorola, Inc. This 
 *  software is provided on an "AS IS" basis and without warranty.
 *
 *  To the maximum extent permitted by applicable law, MOTOROLA 
 *  DISCLAIMS ALL WARRANTIES WHETHER EXPRESS OR IMPLIED, INCLUDING 
 *  IMPLIED WARRANTIES OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR
 *  PURPOSE AND ANY WARRANTY AGAINST INFRINGEMENT WITH REGARD TO THE 
 *  SOFTWARE (INCLUDING ANY MODIFIED VERSIONS THEREOF) AND ANY 
 *  ACCOMPANYING WRITTEN MATERIALS.
 * 
 *  To the maximum extent permitted by applicable law, IN NO EVENT
 *  SHALL MOTOROLA BE LIABLE FOR ANY DAMAGES WHATSOEVER (INCLUDING 
 *  WITHOUT LIMITATION, DAMAGES FOR LOSS OF BUSINESS PROFITS, BUSINESS 
 *  INTERRUPTION, LOSS OF BUSINESS INFORMATION, OR OTHER PECUNIARY
 *  LOSS) ARISING OF THE USE OR INABILITY TO USE THE SOFTWARE.   
 * 
 *  Motorola assumes no responsibility for the maintenance and support
 *  of this software
 ********************************************************************/

/*
 * File:		fec.c
 * Purpose:		Device driver for the MCF5272 FEC
 *
 *      
 */

#include "mcf5272.h"
#include "init.h"
#include "nif.h"
#include "fec.h"

/********************************************************************/
static int
fec_reset(NIF *nif)
{
	MCF5272_IMM *imm = mcf5272_get_immp();

	/* Reset the FEC - equivalent to a hard reset */
	MCF5272_WR_FEC_ECR(imm,MCF5272_FEC_ECR_RESET);
	/* Wait for the reset sequence to complete */
	while (MCF5272_RD_FEC_ECR(imm) & MCF5272_FEC_ECR_RESET) 
		;
	/* Disable all FEC interrupts by clearing the IMR register */
	MCF5272_WR_FEC_IMR(imm,0);
	/* Clear any interrupts by setting all bits in the EIR register */
	MCF5272_WR_FEC_EIR(imm,0xFFFFFFFF);

	/* Set the source address for the controller */
	MCF5272_WR_FEC_MALR(imm,0 
		| (nif->hwa[0] <<24) 
		| (nif->hwa[1] <<16)	
		| (nif->hwa[2] <<8) 
		| (nif->hwa[3] <<0)); 
	MCF5272_WR_FEC_MAUR(imm,0
		| (nif->hwa[4] <<24)
		| (nif->hwa[5] <<16));

	/* Initialize the hash table registers */
	MCF5272_WR_FEC_HTUR(imm,0);
	MCF5272_WR_FEC_HTLR(imm,0);

	/* Set Receive Buffer Size */
	MCF5272_WR_FEC_EMRBR(imm, (uint16)RX_BUFFER_SIZE);

	/* Point to the start of the circular Rx buffer descriptor queue */
	MCF5272_WR_FEC_ERDSR(imm, nbuf_get_start(Rx));

	/* Point to the start of the circular Tx buffer descriptor queue */
	MCF5272_WR_FEC_ETDSR(imm, nbuf_get_start(Tx));

	/* Set the tranceiver interface to MII mode */
	MCF5272_WR_FEC_RCR(imm, 0
		| MCF5272_FEC_RCR_MII_MODE);
	/*	| MCF5272_FEC_RCR_DRT);		/* half duplex */
	
	/* Only operate in half-duplex, no heart beat control */
	MCF5272_WR_FEC_TCR(imm, 0);

	return TRUE;
}

/********************************************************************/
static void 
fec_start(NIF *nif)
{
	MCF5272_IMM *imm = mcf5272_get_immp();
	(void) nif;

	/* Enable FEC */
	MCF5272_WR_FEC_ECR(imm, MCF5272_FEC_ECR_ETHER_EN);

	/* Allow interrupts by setting IMR register */
	MCF5272_WR_FEC_IMR(imm, MCF5272_FEC_IMR_RXFEN);

	/* Indicate that there have been empty receive buffers produced */
	MCF5272_WR_FEC_RDAR(imm,1);
}

/********************************************************************/
static void
fec_stop (NIF *nif)
{
	MCF5272_IMM *imm = mcf5272_get_immp();
	(void) nif;

	/* Set the Graceful Transmit Stop bit */
	MCF5272_WR_FEC_TCR(imm, MCF5272_RD_FEC_TCR(imm)
		| MCF5272_FEC_TCR_GTS);

	/* Wait for the current transmission to complete */
	while ( !(MCF5272_RD_FEC_EIR(imm) & MCF5272_FEC_EIR_GRA))
		;

	/* Clear the GRA event */
	MCF5272_WR_FEC_EIR(imm, MCF5272_FEC_EIR_GRA);

	/* Disable the FEC */
	MCF5272_WR_FEC_ECR(imm,0);

	/* Disable all FEC interrupts by clearing the IMR register */
	MCF5272_WR_FEC_IMR(imm,0);

	/* Clear the GTS bit so frames can be tranmitted when restarted */
	MCF5272_WR_FEC_TCR(imm, MCF5272_RD_FEC_TCR(imm)
		& ~MCF5272_FEC_TCR_GTS);
}

/********************************************************************/
static int
fec_send (NIF *nif, HWA_ADDR_P eth_dest, HWA_ADDR_P eth_src, 
		  uint16 eth_type, NBUF *pNbuf)
{
	MCF5272_IMM *imm = mcf5272_get_immp();

	/* If data length of the frame is too long or short, reject */
	if ((pNbuf->length > nif->mtu) || (pNbuf->length <= 0))
		return FALSE;

	/* Write destination ethernet address */
	pNbuf->data[ETH_HDR_OFFSET + 0] = eth_dest[0];
	pNbuf->data[ETH_HDR_OFFSET + 1] = eth_dest[1];
	pNbuf->data[ETH_HDR_OFFSET + 2] = eth_dest[2];
	pNbuf->data[ETH_HDR_OFFSET + 3] = eth_dest[3];
	pNbuf->data[ETH_HDR_OFFSET + 4] = eth_dest[4];
	pNbuf->data[ETH_HDR_OFFSET + 5] = eth_dest[5];

	/* Write source ethernet address. */
	pNbuf->data[ETH_HDR_OFFSET + 6] = eth_src[0];
	pNbuf->data[ETH_HDR_OFFSET + 7] = eth_src[1];
	pNbuf->data[ETH_HDR_OFFSET + 8] = eth_src[2];
	pNbuf->data[ETH_HDR_OFFSET + 9] = eth_src[3];
	pNbuf->data[ETH_HDR_OFFSET + 10] = eth_src[4];
	pNbuf->data[ETH_HDR_OFFSET + 11] = eth_src[5];

	/* Write frame type */
	pNbuf->data[ETH_HDR_OFFSET + 12] = (uint8)((eth_type & 0xFF00) >> 8);
	pNbuf->data[ETH_HDR_OFFSET + 13] = (uint8)(eth_type & 0x00FF);

	/* Adjust the length by the ethernet header size */
	pNbuf->length += ETH_HDR_SIZE;

	/* Set Frame ready for transmit */
	pNbuf->status |= TX_BD_R;

	/* Mark the buffer as not in use */
	nif->tx_free(pNbuf);

	/* Indicate that there has been a transmit buffer produced */
	MCF5272_WR_FEC_TDAR(imm,1);

	/* The hardware will mark the buffer as empty once it's transmitted */
	return TRUE;
}

/********************************************************************/
static void 
fec_receive(NIF *nif)
{
	NBUF *pNbuf;
	eth_frame_hdr *ethframe;
	MCF5272_IMM *imm = mcf5272_get_immp();

	while( nbuf_rx_next_ready() )	
	{
		/* Get the next buffer received by the FEC */
		pNbuf = nif->rx_alloc();
		if (pNbuf == NULL)
			return;

		/************************************************************
		 The size of the Rx Buffers is set so that every frame addressed
		 to me should fit in one buffer. Therefore, the L bit should be
		 set on every buffer received.
		*************************************************************/
		if (!(pNbuf->status & RX_BD_L))
		{
			do
			{	/* Free the current buffer */
				nif->rx_free(pNbuf);
				/* Get the next buffer of this frame */
				pNbuf = nif->rx_alloc();
				if (pNbuf == NULL)
					return;

				/* Keep going until we find the last one */
			} while (!(pNbuf->status & RX_BD_L));

			/* Free the last buffer of this frame */
			nif->rx_free(pNbuf);
			return;
		}

		/* Point to the ethernet header of this packet */
		ethframe = (eth_frame_hdr *)&pNbuf->data[ETH_HDR_OFFSET];

		/* Make sure the associated protocol is supported */
		if (!nif_protocol_exist(nif,ethframe->type))
		{
			nif->rx_free(pNbuf);
			return;
		}

		/* Adjust the length by the ethernet header size */
		pNbuf->length -= (4 + 6 + 6 + 2);

		/* Call the appriopriate handler */
		nif_protocol_handler(nif, ethframe->type, pNbuf);

		/************************************************************
		 Higher level protocols must save off the data if they need to 
		 keep it around for processing. The buffers are freed here to
		 ensure that the circular buffer descriptor ring is maintained
		 properly.
		*************************************************************/
		nif->rx_free(pNbuf);
	}

	/* Clear interrupt from EIR register */
	MCF5272_WR_FEC_EIR(imm,MCF5272_FEC_EIR_RXF);
}

/********************************************************************/
void
fec_rx_release(NBUF* pNbuf)
{
	MCF5272_IMM *imm = mcf5272_get_immp();

	/* Mark the buffer as empty */
	nbuf_rx_release(pNbuf);

	/* Tell the HW that there has been an empty Rx buffer produced */
	MCF5272_WR_FEC_RDAR(imm,1);
}

/********************************************************************/
int
fec_handler (MCF5272_IMM *imm, NIF *nif)
{	/* This is the Interrupt Service Routine for the FEC receiver */
	
	mcf5272_irq_disable();

	/* Clear FEC RX Event from the Event Register (by writing 1) */
	if (MCF5272_RD_FEC_EIR(imm) & MCF5272_FEC_EIR_RXF)
	{
		/* Clear interrupt from EIR register immediately */
		MCF5272_WR_FEC_EIR(imm,MCF5272_FEC_EIR_RXF);

		fec_receive(nif);
	}

	mcf5272_irq_enable();

	return TRUE;
}

/********************************************************************/
int
fec_init (NIF *nif)
{
	nif_init(nif, "mcf5272_fec");
	nif->hwa_size = 6;
	nif->mtu = ETH_MAX_SIZE;
	nif->reset = fec_reset;
	nif->start = fec_start;
	nif->stop = fec_stop;
	nif->send = fec_send;
	nif->receive = (void *)&fec_receive;
	nif->rx_alloc = nbuf_rx_allocate;
	nif->tx_alloc = nbuf_tx_allocate;
	nif->rx_free = fec_rx_release;
	nif->tx_free = nbuf_tx_release;

	return TRUE;
}

/********************************************************************/
