/*********************************************************************
 *
 * Copyright:
 *	1999-2000 MOTOROLA, INC. All Rights Reserved.  
 *  You are hereby granted a copyright license to use, modify, and
 *  distribute the SOFTWARE so long as this entire notice is
 *  retained without alteration in any modified and/or redistributed
 *  versions, and that such modified versions are clearly identified
 *  as such. No licenses are granted by implication, estoppel or
 *  otherwise under any patents or trademarks of Motorola, Inc. This 
 *  software is provided on an "AS IS" basis and without warranty.
 *
 *  To the maximum extent permitted by applicable law, MOTOROLA 
 *  DISCLAIMS ALL WARRANTIES WHETHER EXPRESS OR IMPLIED, INCLUDING 
 *  IMPLIED WARRANTIES OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR
 *  PURPOSE AND ANY WARRANTY AGAINST INFRINGEMENT WITH REGARD TO THE 
 *  SOFTWARE (INCLUDING ANY MODIFIED VERSIONS THEREOF) AND ANY 
 *  ACCOMPANYING WRITTEN MATERIALS.
 * 
 *  To the maximum extent permitted by applicable law, IN NO EVENT
 *  SHALL MOTOROLA BE LIABLE FOR ANY DAMAGES WHATSOEVER (INCLUDING 
 *  WITHOUT LIMITATION, DAMAGES FOR LOSS OF BUSINESS PROFITS, BUSINESS 
 *  INTERRUPTION, LOSS OF BUSINESS INFORMATION, OR OTHER PECUNIARY
 *  LOSS) ARISING OF THE USE OR INABILITY TO USE THE SOFTWARE.   
 * 
 *  Motorola assumes no responsibility for the maintenance and support
 *  of this software
 ********************************************************************/

/*
 * File:		timer.c
 * Purpose:		routines for accessing integrated timer modules
 *
 * Notes:
 *
 * Modifications:
 *
 */

#include "mcf5272.h"
#include "init.h"
#include "timer.h"

/********************************************************************/

/* The MCF5272 has four general purpose timer modules */
mcf5272_timer timer[4] = \
	{
		{0, 0, 0, 0, 0, 0, 0, 0, 0, 0}, 
		{0, 0, 0, 0, 0, 0, 0, 0, 0, 0}, 
		{0, 0, 0, 0, 0, 0, 0, 0, 0, 0}, 
		{0, 0, 0, 0, 0, 0, 0, 0, 0, 0}
	};

const uint8 TIMER_NETWORK = 1;

/********************************************************************/
uint32
timer_default_isr(MCF5272_IMM *immp, mcf5272_timer *t)
{
	/* Clear the pending event */
	MCF5272_WR_TIMER_TER(immp,t->channel,0
		| MCF5272_TIMER_TER_REF
		| MCF5272_TIMER_TER_CAP);

	t->timeouts++;
	if (t->reference)
		if (--t->reference == 0)
			timer_stop(t->channel);

	return TRUE;
}

/********************************************************************/
uint32
timer_read(uint8 channel)
{
	uint16 now;

	now = MCF5272_RD_TIMER_TCN(timer[channel].immp, channel);
	timer[channel].then = timer[channel].now;
	timer[channel].now = now;
	return now;
}

/********************************************************************/
void
timer_start(uint8 channel)
{
	/* Reset Timer module */
	MCF5272_WR_TIMER_TMR(timer[channel].immp,channel,0);

	/* Clear any pending Timer events */
	MCF5272_WR_TIMER_TER(timer[channel].immp,channel,0
		| MCF5272_TIMER_TER_REF
		| MCF5272_TIMER_TER_CAP);
	
	/* Set Timer Interrupt Control Register */
	MCF5272_WR_SIM_ICR1(timer[channel].immp, 0
		| MCF5272_SIM_ICR_TMR_IL(timer[channel].level, channel));

	/* Reset the counter */
	MCF5272_WR_TIMER_TCN(timer[channel].immp,channel,0);

	/* Write the TRR register */
	MCF5272_WR_TIMER_TRR(timer[channel].immp,channel,timer[channel].trr);

	/* Write the TMR register and start the free-running timer */
	MCF5272_WR_TIMER_TMR(timer[channel].immp,channel,timer[channel].tmr);
}

/********************************************************************/
void
timer_stop(uint8 channel)
{
	/* Get the latest time */
	timer_read(channel);

	/* Disable timer (reset) */
	MCF5272_WR_TIMER_TMR(timer[channel].immp,channel,0);
}

/********************************************************************/
uint32
timer_set_secs(uint8 channel, uint32 secs)
{
	uint32 timeout;

	/* Get the timeout in seconds */
	timeout = (uint32)(((timer[channel].period * 0xFFFF)/1000000000) + 0.5);

	if (timeout == 0)
	{
		timer[channel].reference = 1;
		return FALSE;
	}

	/* Save the reference reached counter value */
	timer[channel].reference = secs/timeout;

	/* Reset the timeout counter */
	timer[channel].timeouts = 0;

	/* Start the timer */
	timer_start(channel);

	return TRUE;
}

/********************************************************************/
uint32
timer_get_reference(uint8 channel)
{	
	return timer[channel].reference;
}

/********************************************************************/
uint32
timer_init(MCF5272_IMM *immp, uint8 channel, float period, 
		   float sysclock, uint32 level)
{
	uint16 prescale;

	if ((channel < 0) || (channel > 3) || (level < 1) || (level > 7))
		return FALSE;

	timer[channel].immp = immp;
	
	/* If no period, disable timer */
	if (!period)
	{
		timer[channel].tmr = 0;
		timer[channel].trr = 0;
		return TRUE;
	}

	/* Set the Reference Reached value to the maximum value */
	timer[channel].trr = 0xFFFF;

	/* Calculate TMR value */
	timer[channel].tmr = (MCF5272_TIMER_TMR_ORI
						| MCF5272_TIMER_TMR_CE_NONE
						| MCF5272_TIMER_TMR_OM
						| MCF5272_TIMER_TMR_FRR
						| MCF5272_TIMER_TMR_RST);

	if (period < (1000/sysclock)) 
	{
		period = 1000/sysclock;
		timer[channel].tmr |= (MCF5272_TIMER_TMR_CLK_MSCLK 
							 | MCF5272_TIMER_TMR_PS(0));
	}
	else if (period > (1000 * 256 * 16)/sysclock)
	{
		period = (1000 * 256 * 16)/sysclock;
		timer[channel].tmr |= (MCF5272_TIMER_TMR_CLK_DIV16 
							 | MCF5272_TIMER_TMR_PS(256));
	}
	else if (period <= (1000 * 256)/sysclock)
	{
		prescale = (uint16)((period * sysclock)/1000);
		timer[channel].tmr |= (MCF5272_TIMER_TMR_CLK_MSCLK 
							 | MCF5272_TIMER_TMR_PS(prescale));
	}
	else /* period <= (1000 * 256 * 16)/sysclock */
	{
		prescale = (uint16)((period * sysclock)/(1000 * 16));
		timer[channel].tmr |= (MCF5272_TIMER_TMR_CLK_DIV16 
							 | MCF5272_TIMER_TMR_PS(prescale));
	}

	/* Save the Timer settings */
	timer[channel].period = period;
	timer[channel].channel = channel;
	timer[channel].level = level;

	return TRUE;
}

/********************************************************************/